/*
 *  stackdump.c
 *
 *  Created by Hans van der Meer on 05-08-08.
 *  Copyright 2008 hvandermeer.com. All rights reserved.
 *
 */

/* stackdump.c -- dump facility for lua C-code. */

#include "lua.h"
#include "lauxlib.h"
#include "lualib.h"
#include "luahm.h"

/* Dump lua stack -- Programming in Lua, 2nd Edition p.225. */

void entryprint(lua_State *L, const int entry, const char *kv)
{
		if (lua_isrealstring(L, entry))
			fprintf(TRACEFILE, "%s%s", kv, lua_tostring(L, entry));
		else if (lua_isnumber(L, entry))
			fprintf(TRACEFILE, "%s" LUA_NUMBER_FMT, kv, lua_tonumber(L, entry));
		else if (lua_isfunction(L, entry))
			fprintf(TRACEFILE, "%sfunction %p", kv, lua_topointer(L, entry));
		else
			fprintf(TRACEFILE, "%s%s", kv, luaL_typename(L, entry));
}

void tableprint(lua_State *L, const int tindex, const char *ttype)
{
	if (!lua_istable(L, tindex))
	{
		fprintf(TRACEFILE, "\tcontents of %s: not a table but %s\n", 
								ttype, luaL_typename(L, tindex));
		return;
	}
	/* must push table to top */
	fprintf(TRACEFILE, "\tcontents of %s:\n", ttype);
	lua_pushvalue(L, tindex);
	int ttindex = lua_gettop(L);
	lua_pushnil(L); /* first key */
	while (lua_next(L, ttindex) != 0)
	{
		/* key at respectively -2 */
		entryprint(L, -2, "\t(key)");
		/* value at -1 */
		entryprint(L, -1, "\t(value)");
		lua_pop(L, 1); /* pop value, keep key for next round */
		fprintf(TRACEFILE, "\n");
	}
	lua_pop(L, 1);	/* remove copied table */
}

void stackdump(lua_State *L,
					const char *mark,		/* print after *** stackdump */
					const bool listtable)	/* show (meta)table contents */
{
	
	fprintf(TRACEFILE, "******************************************** stackdump %s\n", mark);
	int i;
	int top = lua_gettop(L);
	
	for (i = 1; i <= top; ++i)
	{
		int t = lua_type(L, i);
		char *tn = (char *)lua_islightuserdata(L, i) ? 
						"lightuserdata" : (char *)lua_typename(L, t);
		size_t l;
		char *meta = "(no mt)", *metastr = "(no__tostring)";
		char *s;
		
		fprintf(TRACEFILE, "(stack %d/%d %s) ", i, i-top-1, tn);
		switch (t)
		{
			case LUA_TNIL:
				break;
				
			case LUA_TBOOLEAN:
				fprintf(TRACEFILE, "%s\n", (lua_toboolean(L, i) ? "true" : "false"));
				break;
				
			case LUA_TLIGHTUSERDATA:
				fprintf(TRACEFILE, "%p\n", lua_touserdata(L, i));
				break;
				
			case LUA_TNUMBER:
				fprintf(TRACEFILE, "%g\n", lua_tonumber(L, i));
				break;
				
			case LUA_TSTRING:
				s = (char *)lua_tolstring(L, i, &l);
				fprintf(TRACEFILE, "[l=%lu] %s\n", l, s);
				break;
				
			case LUA_TTABLE:
				if (lua_getmetatable(L, i))
				{
					fprintf(TRACEFILE, "(has mt)\n");
					lua_pop(L, 1);
				} else fprintf(TRACEFILE, "(no mt)\n");
				if (listtable)
				{
					tableprint(L, i, "table");
					if (lua_getmetatable(L, i))
					{
						tableprint(L, i, "metatable");
						lua_pop(L, 1);
					}				
				}
				break;
				
			case LUA_TFUNCTION:
				break;
				
			case LUA_TUSERDATA:	   /* print description if tostring function */
				if (lua_getmetatable(L, i))
				{
					meta = "(has mt)";
					if (luaL_callmeta(L, i, "__tostring")) {
						metastr = (char *)lua_tostring(L, -1);
						lua_pop(L, 1);						 /* pop function */
					}
					fprintf(TRACEFILE, "%p %s %s\n", lua_touserdata(L, i), meta, metastr);
					if (listtable) tableprint(L, lua_gettop(L), "metatable");
					lua_pop(L,1);							/* pop metatable */
				} 
				else fprintf(TRACEFILE, "%p %s %s\n", lua_touserdata(L, i), meta, metastr);
				break;
				
			case LUA_TTHREAD:
				fprintf(TRACEFILE, "\n");
				break;
				
			default:
				fprintf(TRACEFILE, "CAN'T HAPPEN typevalue = %d\n", t);
				break;
		}
	}
	fprintf(TRACEFILE, "**************************************** end stackdump %s\n", mark);
}

